"""Unit tests for vizro.models.Action."""

import json
import sys

import dash
import plotly
import pytest
from dash import html
from dash._callback_context import context_value
from dash._utils import AttributeDict
from pydantic import ValidationError

from vizro.actions import export_data
from vizro.models._action._action import Action
from vizro.models.types import capture


@pytest.fixture
def custom_action_function_mock_return(request):
    @capture("action")
    def action_function():
        return request.param

    return action_function


@pytest.fixture
def expected_get_callback_mapping_inputs(request):
    return {
        f'{input["component_id"]}_{input["component_property"]}': dash.State(
            input["component_id"], input["component_property"]
        )
        for input in request.param
    }


@pytest.fixture
def expected_get_callback_mapping_outputs(request):
    return {
        f'{output["component_id"]}_{output["component_property"]}': dash.Output(
            output["component_id"], output["component_property"], allow_duplicate=True
        )
        for output in request.param
    }


@pytest.fixture
def callback_context_set_outputs_grouping(request):
    """Mock dash.callback_context that represents outputs grouping for custom action."""
    outputs = request.param

    outputs_grouping = dict.fromkeys(outputs)
    mock_callback_context = {"outputs_grouping": {"action_finished": None, **outputs_grouping}}
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


@pytest.fixture
def custom_action_build_expected():
    return html.Div(
        children=[],
        id="action_test_action_model_components_div",
    )


class TestActionInstantiation:
    """Tests model instantiation."""

    def test_create_action_mandatory_only(self, test_action_function):
        action = Action(function=test_action_function)

        assert hasattr(action, "id")
        assert action.function == test_action_function
        assert action.inputs == []
        assert action.outputs == []

    def test_create_action_mandatory_and_optional(self, test_action_function):
        inputs = ["component_1.property_A", "component_1.property_B"]
        outputs = ["component_2.property_A", "component_2.property_B"]

        action = Action(function=test_action_function, inputs=inputs, outputs=outputs)

        assert hasattr(action, "id")
        assert action.function == test_action_function
        assert action.inputs == inputs
        assert action.outputs == outputs

    @pytest.mark.parametrize(
        "inputs, outputs",
        [
            ([], []),
            (["component.property"], ["component.property"]),
            (["component.property", "component.property"], ["component.property", "component.property"]),
            (
                ["Component_1.Property_A", "Component_2.Property_B"],
                ["Component_1.Property_A", "Component_2.Property_B"],
            ),
        ],
    )
    def test_inputs_outputs_valid(self, inputs, outputs, test_action_function):
        action = Action(function=test_action_function, inputs=inputs, outputs=outputs)

        assert action.inputs == inputs
        assert action.outputs == outputs

    @pytest.mark.parametrize(
        "inputs",
        [
            "string",
            "component_property",
            "compo-nent.property",
            "component.property_1",
        ],
    )
    def test_inputs_invalid(self, inputs, test_action_function):
        with pytest.raises(ValidationError, match="value is not a valid list"):
            Action(function=test_action_function, inputs=inputs, outputs=[])

    @pytest.mark.parametrize(
        "outputs",
        [
            "string",
            "component_property",
            "compo-nent.property",
            "component.property_1",
        ],
    )
    def test_outputs_invalid(self, outputs, test_action_function):
        with pytest.raises(ValidationError, match="value is not a valid list"):
            Action(function=test_action_function, inputs=[], outputs=outputs)

    @pytest.mark.parametrize("file_format", [None, "csv", "xlsx"])
    def test_export_data_file_format_valid(self, file_format):
        action = Action(id="action_test", function=export_data(file_format=file_format))
        assert action.id == "action_test"
        assert action.inputs == []
        assert action.outputs == []

    def test_export_data_file_format_invalid(self):
        with pytest.raises(
            ValueError, match='Unknown "file_format": invalid_file_format.' ' Known file formats: "csv", "xlsx".'
        ):
            Action(function=export_data(file_format="invalid_file_format"))

    def test_export_data_xlsx_without_required_libs_installed(self, monkeypatch):
        monkeypatch.setitem(sys.modules, "openpyxl", None)
        monkeypatch.setitem(sys.modules, "xlswriter", None)
        # monkeypatch.setattr(
        #     importlib.util,
        #     "find_spec",
        #     lambda arg: None if arg in ["openpyxl", "xlsxwriter"] else importlib.util.find_spec(arg),
        # )

        with pytest.raises(
            ModuleNotFoundError, match="You must install either openpyxl or xlsxwriter to export to xlsx format."
        ):
            Action(function=export_data(file_format="xlsx"))


class TestActionBuild:
    """Tests action build method."""

    def test_custom_action_build(self, test_action_function, custom_action_build_expected):
        action = Action(id="action_test", function=test_action_function)
        result = json.loads(json.dumps(action.build(), cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(custom_action_build_expected, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected


class TestActionPrivateMethods:
    """Test action private methods."""

    def test_get_callback_mapping_no_inputs_no_outputs(self, test_action_function):
        action = Action(id="action_test", function=test_action_function)
        callback_inputs, callback_outputs, action_components = action._get_callback_mapping()
        assert callback_inputs == {
            "trigger": dash.Input({"action_name": "action_test", "type": "action_trigger"}, "data")
        }
        assert callback_outputs == {"action_finished": dash.Output("action_finished", "data")}
        assert action_components == []

    @pytest.mark.parametrize(
        "inputs_and_outputs, expected_get_callback_mapping_inputs, expected_get_callback_mapping_outputs",
        [
            (
                ["component_1.property"],
                [{"component_id": "component_1", "component_property": "property"}],
                [{"component_id": "component_1", "component_property": "property"}],
            ),
            (
                ["component_1.property", "component_2.property"],
                [
                    {"component_id": "component_1", "component_property": "property"},
                    {"component_id": "component_2", "component_property": "property"},
                ],
                [
                    {"component_id": "component_1", "component_property": "property"},
                    {"component_id": "component_2", "component_property": "property"},
                ],
            ),
        ],
        indirect=["expected_get_callback_mapping_inputs", "expected_get_callback_mapping_outputs"],
    )
    def test_get_callback_mapping_with_inputs_and_outputs(  # pylint: disable=too-many-arguments
        self,
        inputs_and_outputs,
        test_action_function,
        expected_get_callback_mapping_inputs,
        expected_get_callback_mapping_outputs,
    ):
        action = Action(
            id="action_test",
            function=test_action_function,
            inputs=inputs_and_outputs,
            outputs=inputs_and_outputs,
        )
        callback_inputs, callback_outputs, action_components = action._get_callback_mapping()
        assert callback_inputs == {
            **expected_get_callback_mapping_inputs,
            "trigger": dash.Input({"action_name": "action_test", "type": "action_trigger"}, "data"),
        }
        assert callback_outputs == {
            **expected_get_callback_mapping_outputs,
            "action_finished": dash.Output("action_finished", "data"),
        }
        assert action_components == []

    @pytest.mark.parametrize(
        "custom_action_function_mock_return, callback_context_set_outputs_grouping, expected_function_return_value",
        [
            # custom action function return value - None
            (None, [], {"action_finished": None}),
            # custom action function return value - single value
            ("new_value", ["component_1_property"], {"action_finished": None, "component_1_property": "new_value"}),
            # custom action function return value - list of values
            (
                ["new_value", "new_value_2"],
                ["component_1_property", "component_2_property"],
                {"action_finished": None, "component_1_property": "new_value", "component_2_property": "new_value_2"},
            ),
            # custom action function return value - tuple
            (
                ("new_value", "new_value_2"),
                ["component_1_property", "component_2_property"],
                {"action_finished": None, "component_1_property": "new_value", "component_2_property": "new_value_2"},
            ),
            # custom action function return value - dictionary
            (
                {"component_1_property": "new_value"},
                ["component_1_property"],
                {"action_finished": None, "component_1_property": "new_value"},
            ),
        ],
        indirect=["custom_action_function_mock_return", "callback_context_set_outputs_grouping"],
    )
    def test_action_callback_function_return_value_valid(
        self, custom_action_function_mock_return, callback_context_set_outputs_grouping, expected_function_return_value
    ):
        action = Action(function=custom_action_function_mock_return())
        result = action._action_callback_function()
        assert result == expected_function_return_value

    @pytest.mark.parametrize(
        "custom_action_function_mock_return, callback_context_set_outputs_grouping",
        [
            (None, ["component_1_property"]),
            ("new_value", []),
            ("new_value", ["component_1_property", "component_2_property"]),
            (["new_value"], []),
            (["new_value"], ["component_1_property", "component_2_property"]),
            (["new_value", "new_value_2"], ["component_1_property"]),
            ({"component_1_property": "new_value"}, []),
            ({"component_1_property": "new_value"}, ["component_1_property", "component_2_property"]),
            ({"component_1_property": "new_value", "component_2_property": "new_value_2"}, ["component_1_property"]),
        ],
        indirect=True,
    )
    def test_action_callback_function_return_value_invalid(
        self, custom_action_function_mock_return, callback_context_set_outputs_grouping
    ):
        action = Action(function=custom_action_function_mock_return())
        with pytest.raises(
            ValueError,
            match="Number of action's returned elements .(.?.)"
            " does not match the number of action's defined outputs .(.?.).",
        ):
            action._action_callback_function()



import importlib.util
import logging
from typing import Any, Dict, List

from dash import Input, Output, State, callback, ctx, html
from pydantic import Field, validator

import vizro.actions
from vizro.managers._model_manager import ModelID
from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call
from vizro.models.types import CapturedCallable

logger = logging.getLogger(__name__)


class Action(VizroBaseModel):
    """Action to be inserted into `actions` of relevant component.

    Args:
        function (CapturedCallable): See [`CapturedCallable`][vizro.models.types.CapturedCallable].
        inputs (List[str]): Inputs in the form `<component_id>.<property>` passed to the action function.
            Defaults to `[]`.
        outputs (List[str]): Outputs in the form `<component_id>.<property>` changed by the action function.
            Defaults to `[]`.
    """

    function: CapturedCallable = Field(..., import_path=vizro.actions)
    inputs: List[str] = Field(
        [],
        description="Inputs in the form `<component_id>.<property>` passed to the action function.",
        regex="^[a-zA-Z0-9_]+[.][a-zA-Z_]+$",
    )
    outputs: List[str] = Field(
        [],
        description="Outputs in the form `<component_id>.<property>` changed by the action function.",
        regex="^[a-zA-Z0-9_]+[.][a-zA-Z_]+$",
    )

    # TODO: Problem: generic Action model shouldn't depend on details of particular actions like export_data.
    # Possible solutions: make a generic mapping of action functions to validation functions or the imports they
    # require, and make the code here look up the appropriate validation using the function as key
    # This could then also involve other validations currently only carried out at run-time in pre-defined actions, such
    # as e.g. checking if the correct arguments have been provided to the file_format in export_data.
    @validator("function")
    def validate_predefined_actions(cls, function):
        if function._function.__name__ == "export_data":
            file_format = function._arguments.get("file_format")
            if file_format not in [None, "csv", "xlsx"]:
                raise ValueError(f'Unknown "file_format": {file_format}.' f' Known file formats: "csv", "xlsx".')
            if file_format == "xlsx":
                if importlib.util.find_spec("openpyxl") is None and importlib.util.find_spec("xlsxwriter") is None:
                    raise ModuleNotFoundError(
                        "You must install either openpyxl or xlsxwriter to export to xlsx format."
                    )
        return function

    @staticmethod
    def _validate_output_number(outputs, return_value):
        return_value_len = (
            1 if not hasattr(return_value, "__len__") or isinstance(return_value, str) else len(return_value)
        )

        # Raising the custom exception if the callback return value length doesn't match the number of defined outputs.
        if len(outputs) != return_value_len:
            raise ValueError(
                f"Number of action's returned elements ({return_value_len}) does not match the number"
                f" of action's defined outputs ({len(outputs)})."
            )

    def _get_callback_mapping(self):
        """Builds callback inputs and outputs for the Action model callback, and returns action required components.

        callback_inputs, and callback_outputs are "dash.State" and "dash.Output" objects made of three parts:
            1. User configured inputs/outputs - for custom actions,
            2. Vizro configured inputs/outputs - for predefined actions,
            3. Hardcoded inputs/outputs - for custom and predefined actions
                (enable callbacks to live inside the Action loop).

        Returns: List of required components (e.g. dcc.Download) for the Action model added to the `Dashboard`
            container. Those components represent the return value of the Action build method.
        """
        from vizro.actions._callback_mapping._get_action_callback_mapping import _get_action_callback_mapping

        callback_inputs: Dict[str, Any] = {
            **_get_action_callback_mapping(action_id=ModelID(str(self.id)), argument="inputs"),
            **{
                f'{input.split(".")[0]}_{input.split(".")[1]}': State(input.split(".")[0], input.split(".")[1])
                for input in self.inputs
            },
            "trigger": Input({"type": "action_trigger", "action_name": self.id}, "data"),
        }

        callback_outputs: Dict[str, Any] = {
            **_get_action_callback_mapping(action_id=ModelID(str(self.id)), argument="outputs"),
            **{
                f'{output.split(".")[0]}_{output.split(".")[1]}': Output(
                    output.split(".")[0], output.split(".")[1], allow_duplicate=True
                )
                for output in self.outputs
            },
            "action_finished": Output("action_finished", "data", allow_duplicate=True),
        }

        action_components = _get_action_callback_mapping(action_id=ModelID(str(self.id)), argument="components")

        return callback_inputs, callback_outputs, action_components

    def _action_callback_function(self, **inputs: Dict[str, Any]) -> Dict[str, Any]:
        logger.debug("=============== ACTION ===============")
        logger.debug(f'Action ID: "{self.id}"')
        logger.debug(f'Action name: "{self.function._function.__name__}"')
        logger.debug(f"Action inputs: {inputs}")

        # Invoking the action's function
        return_value = self.function(**inputs) or {}

        # Action callback outputs
        outputs = list(ctx.outputs_grouping.keys())
        outputs.remove("action_finished")

        # Validate number of outputs
        self._validate_output_number(
            outputs=outputs,
            return_value=return_value,
        )

        # If return_value is a single element, ensure return_value is a list
        if not isinstance(return_value, (list, tuple, dict)):
            return_value = [return_value]
        if isinstance(return_value, dict):
            return {"action_finished": None, **return_value}

        return {"action_finished": None, **dict(zip(outputs, return_value))}

    @_log_call
    def build(self):
        """Builds a callback for the Action model and returns required components for the callback.

        Returns:
            List of required components (e.g. dcc.Download) for the Action model added to the `Dashboard` container.
        """
        callback_inputs, callback_outputs, action_components = self._get_callback_mapping()

        logger.debug(
            f"Creating Callback mapping for Action ID {self.id} with "
            f"function name: {self.function._function.__name__}"
        )
        logger.debug("---------- INPUTS ----------")
        for name, object in callback_inputs.items():
            logger.debug(f"--> {name}: {object}")
        logger.debug("---------- OUTPUTS ---------")
        for name, object in callback_outputs.items():
            logger.debug(f"--> {name}: {object}")
        logger.debug("============================")

        @callback(output=callback_outputs, inputs=callback_inputs, prevent_initial_call=True)
        def callback_wrapper(trigger: None, **inputs: Dict[str, Any]) -> Dict[str, Any]:
            return self._action_callback_function(**inputs)

        # return action_components
        return html.Div(
            children=action_components,
            id=f"{self.id}_action_model_components_div",
        )



from typing import Optional

import dash
import dash_bootstrap_components as dbc
from dash import html
from pydantic import validator

from vizro._constants import ACCORDION_DEFAULT_TITLE
from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call
from vizro.models._navigation.navigation import _validate_pages
from vizro.models.types import NavigationPagesType


class Accordion(VizroBaseModel):
    """Accordion to be used in Navigation Panel of Dashboard.

    Args:
        pages (Optional[NavigationPagesType]): See [NavigationPagesType][vizro.models.types.NavigationPagesType].
            Defaults to `None`.
    """

    pages: Optional[NavigationPagesType] = None

    # validators
    _validate_pages = validator("pages", allow_reuse=True, always=True)(_validate_pages)

    @_log_call
    def build(self, *, active_page_id=None):
        return self._create_accordion(active_page_id=active_page_id)

    def _create_accordion_buttons(self, pages, active_page_id):
        """Creates a button for each provided page that is registered."""
        accordion_buttons = []
        for page_id in pages:
            try:
                page = dash.page_registry[page_id]
            except KeyError as exc:
                raise KeyError(
                    f"Page with ID {page_id} cannot be found. Please add the page to `Dashboard.pages`"
                ) from exc
            accordion_buttons.append(
                dbc.Button(
                    children=[page["name"]],
                    key=page["relative_path"],
                    className="accordion-item-button",
                    active=page_id == active_page_id,
                    href=page["relative_path"],
                )
            )
        return accordion_buttons

    def _create_accordion_item(self, accordion_buttons, title=ACCORDION_DEFAULT_TITLE):
        """Creates an accordion item for each sub-group of pages."""
        return dbc.AccordionItem(
            children=accordion_buttons,
            title=title.upper(),
            class_name="accordion_item",
        )

    def _get_accordion_container(self, accordion_items, accordion_buttons):
        # Return no container if there is only one page in the dashboard or no pages exist
        if (len(accordion_buttons) == len(accordion_items) == 1) or not accordion_buttons:
            return None

        return html.Div(
            children=[
                dbc.Accordion(
                    id=self.id,
                    children=accordion_items,
                    class_name="accordion",
                    persistence=True,
                    persistence_type="session",
                    always_open=True,
                ),
                html.Hr(),
            ],
            className="nav_panel",
            id=f"{self.id}_outer",
        )

    def _create_accordion(self, active_page_id):
        """Creates a custom accordion only with user-provided pages."""
        accordion_items = []
        if isinstance(self.pages, dict):
            for page_group, page_members in self.pages.items():
                accordion_buttons = self._create_accordion_buttons(pages=page_members, active_page_id=active_page_id)
                accordion_items.append(
                    self._create_accordion_item(accordion_buttons=accordion_buttons, title=page_group)
                )

        if isinstance(self.pages, list):
            accordion_buttons = self._create_accordion_buttons(pages=self.pages, active_page_id=active_page_id)
            accordion_items.append(self._create_accordion_item(accordion_buttons=accordion_buttons))
        return self._get_accordion_container(accordion_items=accordion_items, accordion_buttons=accordion_buttons)
